from .data_group import Group
from .exceptions import GroupNotFoundError, GroupAlreadyExisit
from pathlib import Path


class Repo:
    """
    Mock数据仓库
    """
    def __init__(self, repo_path):
        self.path = Path(repo_path)
        self.groups = {}
        self.activated_group = None

    def activate(self, name):
        group = self.groups.get(name)
        if not group:
            raise GroupNotFoundError(f'Group {name} not found')
        self.activated_group = group

    def deactivate(self, name):
        if self.activated_group and self.activated_group.name == name:
            self.activated_group = None

    def reset(self):
        self.activated_group = None

    def create_group(self, name):
        if self.groups.get(name):
            raise GroupAlreadyExisit(f'Group {name} already exisit')
        self.groups[name] = Group(name)
        self.groups[name].save()

    def delete_group(self, name):
        group = self.groups.pop(name)
        if not group:
            raise GroupNotFoundError(f'Group {name} not found')
        group.remove()

    def load(self):
        for sub_file in self.path.iterdir():
            if not sub_file.is_dir():
                continue
            group = Group(sub_file)
            group.load()
            self.groups[group.name] = group 

    def save(self):
        pass



from pathlib import Path
import json
import codecs


def dict_to_obj(d, o):
    return o.__dict__.update(d)

def obj_to_dict(o, d):
    pass


class Request:
    
    def __init__(self):
        self.url = None        
        self.version = None
        self.method = None
        self.scheme = None
        self.host = None
        self.path = None
        self.qurey = None
        self.headers = {}
        self.data = None
        self._path: Path = None

    def save(self):
        pass

    def load(self):
        req = self._path/'request.json'
        if req.exists:
            with codecs.open(str(req)) as f:
                req_json_obj = json.loads(f.read())
                dict_to_obj(req_json_obj, self)
        req_data_json = self._path/'request_data.json'
        req_data_txt = self._path/'request_data.bin'
        if req_data_json.exists():
            with codecs.open(str(req_data_json)) as f:
                req_data_obj = json.loads(f.read())
                self.data = {}
                dict_to_obj(req_data_obj, self.data)
        elif req_data_txt.exists():
            with codecs.open(str(req_data_txt)) as f:
                req_data_obj = json.loads(f.read())
                self.data = {}
                dict_to_obj(req_data_obj, self.data)

        


class Response:
    
    def __init__(self):
        self.status = -1
        self.headers = {}
        self.data = None

    def save(self):
        pass

    def load(self):
        pass


class Flow:
    
    def __init__(self):
        self.id = None
        self.path: Path = None
        self.request: Request = None
        self.response: Response = None
        self.timings = None

    @classmethod
    def from_path(cls, path):
        instance = cls()
        instance.path = path
        instance.load()
        return instance

    def save(self):
        pass
    
    def load(self):
        self.request = Request()
        self.request._path = self.path
        self.request.load()



from mitmproxy import http
from lyrebird import application
from lyrebird import log

"""
Script for mitmdump

Redirect request from proxy server to mock server
"""

_logger = log.get_logger()


def to_mock_server(flow: http.HTTPFlow):
    conf = application.config
    # mock path 为/mock开头加上原始url
    flow.request.path = '/mock/' + flow.request.url
    # mock scheme 统一为http
    flow.request.scheme = 'http'
    # mock server port
    flow.request.port = conf.get('mock.port')
    # mock server ip
    flow.request.host = conf.get('mock.host', 'localhost')
    # device real ip
    flow.request.headers['lyrebird.device.ip'] = flow.client_conn.address[0]
    _logger.info('Redirect-> %s' % flow.request.url[:100])


def request(flow: http.HTTPFlow):
    conf = application.config
    _logger.info(flow.request.url[:100])
    filters = conf.get('proxy.filters')
    if not filters:
        to_mock_server(flow)
        return
    for _filter in filters:
        if _filter in flow.request.host:
            to_mock_server(flow)
            break



class DataManagerError(Exception):
    pass


class GroupNotFoundError(DataManagerError):
    pass


class GroupAlreadyExisit(DataManagerError):
    pass


class SerializeTargetNotExisit(DataManagerError):
    pass



from flask_sqlalchemy import SQLAlchemy

class DataBase:
    def __init__(self, app):
        self._db = SQLAlchemy(app)
        self._db.init_app(app)

    @property
    def model(self):
        return self._db.Model

    @property
    def session(self):
        return self._db.session

    def create_all(self):
        self._db.create_all()



import json
from pathlib import Path
from .exceptions import SerializeTargetNotExisit


class Serializable:

    def __init__(self, path):
        self.path = path
    
    def save(self):
        pass

    def load(self):
        pass


class SerializableDir(Serializable):

    pass


class SerializableJson(Serializable):

    pass


def from_path(target_path):
    target = Path(target_path)
    if target.is_dir():
        pass
    elif target.is_file():
        SerializableJson(target_path)
    else:
        SerializableDir(target_path)
    if not target.exists():
        raise SerializeTargetNotExisit(f'Target {target_path} not exisit')




from lyrebird.mock import context
from lyrebird import application
from lyrebird.mock.filesystem import DataGroupConfBuilder
from flask import Blueprint, request, jsonify, abort
from flask_restful import Resource, Api
import json
import traceback
import time
from urllib.parse import urlparse, unquote
from lyrebird.mock.filesystem import DataHelper
from .. import plugin_manager


"""
实现控制API
"""

api = Blueprint('api', __name__, url_prefix='/api')
api_source = Api(api)


class Status(Resource):

    def get(self):
        conf = context.application.conf
        return context.make_ok_response(
            **{'ip':conf.get('ip', 'unknown ip'), 
            'mock.port':conf['mock.port'], 
            'proxy.port':conf['proxy.port']})


class Flow(Resource):
    """
    当前请求单条数据
    """
    def get(self, id):
        for item in context.application.cache.items():
            if item['id'] == id:
                return jsonify(item)
        return abort(400, 'Request not found')


class FlowList(Resource):
    """
    当前请求列表
    """
    def get(self):
        all_items = context.application.cache.items()[::-1]
        req_list = []
        for item in all_items:
            info = dict()
            info['id'] = item['id']
            info['time'] = item['time']
            info['response-time'] = item['response-time']
            info['request'] = dict()
            info['request']['url'] = item['request']['url']
            info['request']['method'] = item['request']['method']
            info['response'] = dict()
            info['response']['code'] = item['response']['code']
            info['response']['mock'] = item['response']['headers'].get('lyrebird', 'proxy')
            req_list.append(info)
        return jsonify(req_list)

    def delete(self):
        _ids = request.form.getlist('ids')
        if _ids:
            context.application.cache.delete_by_ids(_ids)
        else:
            context.application.cache.clear()
        context.application.socket_io.emit('action', 'delete flow log')
        return context.make_ok_response()

    def post(self):
        _ids = request.json.get('ids')
        record_items = []
        for _id in _ids:
            for item in context.application.cache.items():
                if _id == item['id']:
                    record_items.append(item)
                    break
        current_group = context.application.data_manager.current_data_group
        for item in record_items:
            current_group.add_data_and_filter(item)
        return context.make_ok_response()


class MockGroup(Resource):
    """
    mock数据组
    """
    def get(self, group=None):
        if not group:
            return jsonify([group_name for group_name in context.application.data_manager.data_groups])

        _group = context.application.data_manager.data_groups.get(group)
        if _group:
            resp_obj = dict()
            resp_obj['data'] = [[k, _group.data_dict[k].path] for k in _group.data_dict]
            resp_obj['conf'] = _group.conf
            return jsonify(resp_obj)
        return context.make_fail_response('Not found group')

    def post(self):
        name = request.form.get('name')
        data = request.form.get('data')
        origin_name = request.form.get('origin_name')
        if not name or not data:
            return abort('请设置form：name data')
        if origin_name == '':
            # todo json exception and origin_name exception
            context.application.data_manager.add_group(name, json.loads(data))
        else:
            # todo move to update method
            context.application.data_manager.update_group(origin_name, name, json.loads(data))
        return context.make_ok_response()

    def put(self):
        # todo
        return 'MockGroup UPDATE'

    def delete(self, group):
        if group in context.application.data_manager.data_groups:
            _group = context.application.data_manager.data_groups.pop(group)
            _group.delete()
            return context.make_ok_response()
        else:
            return context.make_fail_response('Not found group')


class MockData(Resource):
    """
    mock数据
    """
    def get(self, group=None, data=None):
        if not group or not data:
            return abort(400, '请设置参数 group name')
        _group = context.application.data_manager.data_groups.get(group)
        if _group and data in _group.data_dict:
            _data = _group.data_dict[data]
            _data.read_file()
            return jsonify(_data.json_data)
        return context.make_fail_response('Data not found')

    def put(self, group=None, data=None):
        return 'MockData UPDATE'

    def post(self, group=None, data=None):
        group = request.form.get('group')
        name = request.form.get('name')
        req = request.form.get('req')
        req_data = request.form.get('req_data')
        resp = request.form.get('resp')
        resp_data = request.form.get('resp_data')
        origin_name = request.form.get('origin_name')
        if not group or not name:
            return context.make_fail_response('请设置form：group & name')
        _group = context.application.data_manager.data_groups.get(group)
        if _group:
            data = DataHelper.to_dict(req, req_data, resp, resp_data)
            if origin_name == '' or origin_name is None:
                _group.add_data(name, data)
            else:
                _group.update_data(origin_name, name, data)
            return context.make_ok_response()
        else:
            return context.make_fail_response('Group not found')

    def delete(self, group=None, data=None):
        if not group or not data:
            return abort(400, '请设置参数 group name')
        _group = context.application.data_manager.data_groups.get(group)
        if _group and data in _group.data_dict:
            data = _group.data_dict[data]
            data.delete()
            _group.scan()
            return context.make_ok_response()
        return context.make_fail_response('Data not found')


class ActivatedMockGroup(Resource):

    def get(self):
        _current_data_group = context.application.data_manager.current_data_group
        if _current_data_group:
            return jsonify({'name': _current_data_group.name})
        else:
            return jsonify({'name': None})

    def put(self, group=None, action=None):
        if action == 'activate':
            result = context.application.data_manager.set_current_data_group(group)
        else:
            result = context.application.data_manager.set_current_data_group(None)
        if result:
            return context.make_ok_response()
        else:
            return context.make_fail_response("Activate group fail. Not found! group=%s action=%s" % (group, action))


class Conf(Resource):
    """
    Lyrebird 及 插件 配置文件获取和修改
    """

    def get(self, plugin_name):
        if plugin_name == 'lyrebird':
            return jsonify(application.config.raw())
        else:
            return jsonify(plugin_manager.get_conf(plugin_name))

    def put(self, plugin_name):
        try:
            if plugin_name == 'lyrebird':
                context.application.conf = request.get_json()
                context.application.save()
                return context.make_ok_response()
            else:
                plugin_manager.set_conf(plugin_name, request.get_json())
                return context.make_ok_response()
        except Exception as e:
            return context.make_fail_response(str(e))


class ResetConf(Resource):
    """
    Lyrebird 及 插件 配置文件重置
    """
    def put(self, plugin_name):
        try:
            if plugin_name == 'lyrebird':
                return context.make_fail_response('暂未开放')
            else:
                plugin_manager.set_default_conf(plugin_name)
                return context.make_ok_response()
        except Exception as e:
            return context.make_fail_response(str(e))


class WorkMode(Resource):

    def put(self, mode=None):
        if context.Mode.contains(mode):
            context.application.work_mode = mode
            return context.make_ok_response()
        else:
            return context.make_fail_response(f'Unknown mode: {mode}')

    def get(self):
        return jsonify({'mode': context.application.work_mode})

api_source.add_resource(Status, '/status')
api_source.add_resource(Flow, '/flow/<string:id>')
api_source.add_resource(FlowList, '/flow')
api_source.add_resource(MockGroup, '/mock', '/mock/<string:group>')
api_source.add_resource(MockData, '/mock/<string:group>/data', '/mock/<string:group>/data/<string:data>')
api_source.add_resource(ActivatedMockGroup, '/mock/activated', '/mock/<string:group>/<string:action>')
api_source.add_resource(Conf, '/conf/<string:plugin_name>')
api_source.add_resource(ResetConf, '/conf/<string:plugin_name>/reset')
api_source.add_resource(WorkMode, '/mode', '/mode/<string:mode>')


from lyrebird.mock import context


class MockHandler:
    """
    根据当前设置数据组的匹配条件,查找对应的mock数据。
    如果没有找到匹配的数据则交由下一个处理器处理。

    """
    def handle(self, handler_context):
        data_group = context.application.data_manager.current_data_group
        if data_group:
            handler_context.response = data_group.get_response(handler_context.get_origin_url(),
                                                               handler_context.request.data)



"""
Event bus

Worked as a backgrund thread
Run events handler and background task worker
"""
from queue import Queue
from concurrent.futures import ThreadPoolExecutor
import traceback
import inspect
from lyrebird.base_server import ThreadServer
from lyrebird import application


class Event:
    """
    Event bus inner class
    """
    def __init__(self, channel, message):
        self.channel = channel 
        self.message = message
        if isinstance(message, dict):
            message['channel'] = channel


class EventServer(ThreadServer):
    
    def __init__(self):
        super().__init__()
        self.event_queue = Queue()
        self.state = {}
        self.pubsub_channels = {}
        # channel name is 'any'. For linstening all channel's message
        self.any_channel = []
        self.broadcast_executor = ThreadPoolExecutor(max_workers=10, thread_name_prefix='event-broadcast')

    def broadcast_handler(self, callback_fn, event, args, kwargs):
        try:
            if kwargs.get('event', False):
                callback_fn(event)
            else:
                callback_fn(event.message)
        except Exception:
            # TODO handle exceptioins and send to event bus
            traceback.print_exc()

    def run(self):
        while self.running:
            try:
                e = self.event_queue.get()
                callback_fn_list = self.pubsub_channels.get(e.channel)
                if callback_fn_list:
                    for callback_fn, args, kwargs in callback_fn_list:
                        self.broadcast_executor.submit(self.broadcast_handler, callback_fn, e, args, kwargs)
                for callback_fn, args, kwargs in self.any_channel:
                    self.broadcast_executor.submit(self.broadcast_handler, callback_fn, e, args, kwargs)
            except Exception:
                # empty event
                traceback.print_exc()

    def stop(self):
        super().stop()
        self.publish('any', 'stop')

    def publish(self, channel, message, state=False, *args, **kwargs):
        """
        publish message
        if state is true, message will be keep as state

        """
        self.event_queue.put(Event(channel, message))
        if state:
            self.state[channel] = message


    def subscribe(self, channel, callback_fn, *args, **kwargs):
        """
        Subscribe channel with a callback function
        That function will be called when a new message was published into it's channel

        kwargs:
            event=False receiver gets a message dict
            event=True receiver gets an Event object
        """
        if channel == 'any':
            self.any_channel.append([callback_fn, args, kwargs])
        else:
            callback_fn_list = self.pubsub_channels.setdefault(channel, [])
            callback_fn_list.append([callback_fn, args, kwargs])


    def unsubscribe(self, channel, target_callback_fn, *args, **kwargs):
        """
        Unsubscribe callback function from channel
        """
        if channel == 'any':
            for any_channel_fn, *_ in self.any_channel:
                if target_callback_fn == any_channel_fn:
                    self.any_channel.remove([target_callback_fn, *_])
        else:
            callback_fn_list = self.pubsub_channels.get(channel)
            for callback_fn, *_ in callback_fn_list:
                if target_callback_fn == callback_fn:
                    callback_fn_list.remove([target_callback_fn, *_])


class CustomEventReceiver:
    """
    Event Receiver

    Decorator for plugin developer
    useaeg:
        event = CustomEventReceiver()

        @event('flow')
        def on_message(data):
            pass
    """
    def __init__(self):
        self.listeners = []

    def __call__(self, channel, object=False, *args, **kw):
        def func(origin_func):
            self.listeners.append(dict(channel=channel, func=origin_func, object=object))
            return origin_func
        return func

    def register(self, event_bus):
        for listener in self.listeners:
            event_bus.subscribe(listener['channel'], listener['func'])

    def unregister(self, event_bus):
        for listener in self.listeners:
            event_bus.unsubscribe(listener['channel'], listener['func'])

    def publish(self, channel, message, *args, **kwargs):
        application.server['event'].publish(channel, message, *args, **kwargs)

    def alert(self, channel, message, *args, **kwargs):
        stack = inspect.stack()
        script_name = stack[1].filename
        script_name = script_name[script_name.rfind('/') + 1:]
        function_name = stack[1].function
        if isinstance(message, dict):
            message['script_name'] = script_name
            message['function_name'] = function_name
        application.server['event'].publish(channel, message, *args, **kwargs)



import pkg_resources
from . import handlers
from collections import OrderedDict, namedtuple
import traceback
from jinja2 import Environment, PackageLoader
from flask import send_from_directory
import sys
import os
import inspect
import codecs
import json
from types import FunctionType
from pathlib import Path
from . import context
from lyrebird import log

"""
Plugin manager
"""

logger = log.get_logger()


ROOT_DIR = Path('~/.lyrebird').expanduser()
PLUGINS_DIR = ROOT_DIR/'plugins'

DATA_HANDLER_ENTRY_POINT = 'lyrebird_data_handler'
WEB_ENTRY_POINT = 'lyrebird_web'

inner_handler = OrderedDict()
data_handler_plugins = OrderedDict()
web_plugins = OrderedDict()


class Rule:
    """
    视图路由规则
    
    """
    def __init__(self, rule, endpotion, view_func=None, **options):
        self.rule = rule
        self.endpoint = endpotion
        self.view_func = view_func
        self.options = options

    def __str__(self):
        return f'Rule {self.rule} func={self.view_func}'


class EventRule:
    """
    socket-io事件规则
    """
    def __init__(self, event, handler, namespace):
        self.event = event
        self.handler = handler
        self.namespace = namespace

    def __str__(self):
        return f'Event rule {self.event} handler={self.handler}'


class PluginView:
    """
    插件视图基类
    
    """
    def __init__(self):
        self.env = Environment()
        self.rules = []
        self.event_rules = []
        self.static_folder = None
        self.template_folder = None

    def add_url_rule(self, rule, endpoint=None, view_func=None, **options):
        self.rules.append(Rule(rule, endpoint, view_func, **options))

    def on_event(self, event, handler, namespace):
        self.event_rules.append(EventRule(event, handler, namespace))

    def on_create(self):
        """
        初始生命周期
        通过在该函数内调用'set_template_root'指定前端模板目录
        通过在该函数内调用'set_static_root'指定前端静态文件目录
        通过在该函数内调用'add_url_rule'添加处理请求的路由，插件的主页面要使用'/'路径
        """
        pass

    def after_on_create(self):
        if not self.static_folder:
            self.set_static_root()
        if not self.template_folder:
            self.set_template_root()

    def _get_current_package_name(self):
        """
        通过class判断包名
        :return: 
        """
        module_name = self.__class__.__module__
        if '.' in module_name:
            return module_name.split('.')[0]
        return module_name

    def get_package_file_path(self, file_path, package_name=None):
        if not package_name:
            package_name = self._get_current_package_name()
        target_module = sys.modules.get(package_name)
        package_path = os.path.dirname(inspect.getsourcefile(target_module))
        return os.path.abspath(os.path.join(package_path, file_path))

    def set_template_root(self, package_name=None, templates_path='templates'):
        """
        设置html模板根目录. 默认在 [插件包]/templates 下
        
        :param package_name: 包名，不指定时默认读取当前插件包 
        :param templates_path: 模板路径， 默认为templates目录
        """
        if not package_name:
            package_name = self._get_current_package_name()
        self.env.loader = PackageLoader(package_name, package_path=templates_path)
        target_module = sys.modules.get(package_name)
        package_path = os.path.dirname(target_module.__file__)
        self.template_folder = os.path.abspath(os.path.join(package_path, templates_path))

    def set_static_root(self, package_name=None, static_path='static'):
        """
        设置静态文件(css、js等)目录. 默认在 [插件包]/static 下
        
        :param package_name: 包名，不指定时默认读取当前插件包 
        :param static_path: 静态文件目录， 默认为static目录
        """
        if not package_name:
            package_name = self._get_current_package_name()
        target_module = sys.modules.get(package_name)
        package_path = os.path.dirname(target_module.__file__)
        self.static_folder = os.path.abspath(os.path.join(package_path, static_path))
        self.add_url_rule('/static/<path:filename>', view_func=self.send_static_file)

    def render_template(self, template, *args, **kwargs):
        """
        渲染模板，支持jinja2语法
        
        :param template: 模板文件名 
        :param args: 参数数组
        :param kwargs: 参数字典
        :return: 返回渲染完成的html文本
        """
        return self.env.get_template(template).render(*args, **kwargs)

    def send_static_file(self, filename):
        """
        响应静态文件
        收到请求后，按照filename在静态文件目录下查找，并返回静态文件
        
        :param filename: 
        :return: 
        """
        # todo cache_timeout
        return send_from_directory(self.static_folder, filename)

    def send_template_file(self, filename):
        template_file_path = os.path.abspath(os.path.join(self.template_folder, filename))
        return codecs.open(template_file_path, 'r', 'utf-8').read()

    def default_conf(self):
        """
        设置默认的 conf.json
        
        :return: 返回 conf.json 内容
        """
        # todo plugin conf 应该以对象方式读取存储 并且对外只读，特定方法支持写
        return {}


def _load_data_handler_plugin():
    data_handler_plugins.clear()
    for ep in pkg_resources.iter_entry_points(DATA_HANDLER_ENTRY_POINT):
        try:
            handler_class = ep.load()
            if not isinstance(handler_class, type):
                logger.error(f'Load plugin {ep} error. Entry point is not a class')
                continue
            if not hasattr(handler_class, 'handle'):
                logger.error(f'Load plugin {ep} error. Not found function - "handle"')
                continue
            handler = handler_class()
            data_handler_plugins[ep.name] = handler
        except Exception:
            logger.error(f'Load plugin {ep} error')
            traceback.print_exc()


def get_change_response_plugins():
    need_change_resp_handlers = []
    for plugin in data_handler_plugins.items():
        if hasattr(plugin, 'change_response'):
            if isinstance(plugin.change_response, bool) and plugin.change_response:
                need_change_resp_handlers.append(plugin)
            elif isinstance(plugin.change_response, FunctionType) and plugin.change_response():
                need_change_resp_handlers.append(plugin)
    return need_change_resp_handlers


def _check_data_handlers_response_conflict():
    need_change_resp_handlers = get_change_response_plugins()
    if len(need_change_resp_handlers) > 1:
        logger.error(f'More than one plugin will modify response. Please check those plugins {need_change_resp_handlers}')


def _load_web_plugin():
    web_plugins.clear()
    for ep in pkg_resources.iter_entry_points(WEB_ENTRY_POINT):
        try:
            web_class = ep.load()
        except Exception:
            logger.error(f'Load plugin fail. {ep}')
            traceback.print_exc()
            continue
        if not isinstance(web_class, type):
            logger.error(f'Load plugin {ep} error. Entry point is not a class')
            continue
        web_plugin = web_class()
        if ep.name in web_plugins:
            logger.error(f'Plugin {ep} duplicate name with {web_plugins[ep.name]}')
            continue
        if not hasattr(web_class, 'on_create'):
            logger.error(f'Load plugin {ep} error. Not found function - "on_create"')
            continue
        if not hasattr(web_class, 'index'):
            logger.error(f'Load plugin {ep} error. Not found function - "index"')
            continue
        try:
            web_plugin.on_create()
            web_plugin.after_on_create()
        except Exception:
            logger.error(f'Plugin {ep} on create', traceback.format_exc())
            continue
        web_plugins[ep.name] = web_plugin


def _load_inner_plugin():
    inner_handler.update(handlers.get_inner_handlers())


def load():
    _load_inner_plugin()
    _load_data_handler_plugin()
    _check_data_handlers_response_conflict()
    _load_web_plugin()


def add_view_to_blueprint(blueprint):
    if len(web_plugins) == 0:
        return
    for plugin_name in web_plugins:
        plugin = web_plugins[plugin_name]
        try:
            for rule in plugin.rules:
                # 去掉自定义路由前面的"/"
                if rule.rule.startswith('/'):
                    rule.rule = rule.rule[1:]
                # 为了防止重名，拼接endpoint
                if not rule.endpoint:
                    rule.endpoint = plugin_name + '_' + rule.view_func.__name__
                # 注册视图
                blueprint.add_url_rule(
                    f'/plugin/{plugin_name}/{rule.rule}',
                    rule.endpoint,
                    view_func=rule.view_func,
                    **rule.options
                )
        except Exception:
            logger.error(f'Can not add rule {rule}')
            traceback.print_exc()


def add_event_rules(socket_io):
    if len(web_plugins) == 0:
        return
    for plugin_name in web_plugins:
        plugin = web_plugins[plugin_name]
        try:
            for event_rule in plugin.event_rules:
                socket_io.on_event(event_rule.event, event_rule.handler, event_rule.namespace)
        except Exception:
            logger.error(f'Can not add event rule {event_rule}')
            traceback.print_exc()


def caller_info(index=1):
    stack = inspect.stack()
    caller_module = inspect.getmodule(stack[index].frame)
    caller_module_name = caller_module.__name__
    if caller_module_name.find('.') > 0:
        caller_top_module_name = caller_module_name.split('.')[0]
    else:
        caller_top_module_name = caller_module_name
    CallerInfo = namedtuple('CallerInfo', 'top_module_name module_name')
    return CallerInfo(module_name=caller_module_name, top_module_name=caller_top_module_name)


class PluginNotFoundError(Exception):
    pass


def _get_plugin_conf_path(name):
    plugin = web_plugins.get(name)
    if not plugin:
        raise PluginNotFoundError('Plugin %s not fonud' % name)
    if '.' in plugin.__module__:
        top_module_name = plugin.__module__.split('.')[0]
    else:
        top_module_name = plugin.__module__
    plugin_storage_dir = PLUGINS_DIR/top_module_name
    if not plugin_storage_dir.exists():
        plugin_storage_dir.mkdir()
    return plugin_storage_dir/'conf.json'


def get_conf(name):
    conf_path = _get_plugin_conf_path(name)
    if not conf_path.exists():
        set_default_conf(name)
    with codecs.open(conf_path, 'r' , 'utf-8') as f:
        return json.load(f)


def set_conf(name, conf):
    conf_path = _get_plugin_conf_path(name)
    if not conf_path.exists():
        conf_path.touch()
    with codecs.open(conf_path, 'w', 'utf-8') as f:
        f.write(json.dumps(conf, ensure_ascii=False, indent=4))
    context.application.event_bus.publish('config', dict(name=name, action='update'))

def set_default_conf(name):
    plugin = web_plugins[name]
    set_conf(name, plugin.default_conf())
    context.application.event_bus.publish('config', dict(name=name, action='update'))



import os
from flask import render_template, Blueprint
import traceback
from lyrebird.mock import context
from lyrebird import log
from ... import version
from .. import plugin_manager
from bs4 import BeautifulSoup
import datetime
from ..reporter import report_handler


ui = Blueprint('ui', __name__, url_prefix='/ui', template_folder='../templates', static_folder='../static')

logger = log.get_logger()

def render_with_plugin(template_name_or_list, **context):
    web_plugins = []
    icon = ''
    for web_plugin_name in plugin_manager.web_plugins:
        try:
            obj = plugin_manager.web_plugins[web_plugin_name]
            if hasattr(obj, "index") and not obj.index():
                continue
            if hasattr(obj, "get_icon"):
                icon = obj.get_icon()
            if hasattr(obj, 'get_title'):
                title = obj.get_title()
            else:
                title = web_plugin_name.capitalize()
            web_plugins.append({'name': web_plugin_name, 'icon': icon, 'title': title})
        except Exception:
            logger.error(f'Load web plugin error. {traceback.format_exc()}')
    return render_template(template_name_or_list, web_plugins=web_plugins, version=version.VERSION, **context)


@ui.route('/')
def index():
    report_handler.page_in('inspector')
    return render_with_plugin('inspector.html')


@ui.route('/data_manager')
def data_manager():
    report_handler.page_in('data_manager')
    return render_with_plugin('data_manager.html')


@ui.route('/plugin/base/<string:name>')
def plugin_base(name):
    report_handler.page_in(name)
    plugin = plugin_manager.web_plugins.get(name)
    if not plugin:
        return "Plugin not found"
    
    web_content = plugin.index()
    soup = BeautifulSoup(web_content, 'html.parser')

    # set all javascripts into script block
    all_scripts = []
    for javascript_tag in soup.find_all('script'):
        all_scripts.append(javascript_tag)
        javascript_tag.extract()
    # set all css link into header
    all_css = []
    for css_tag in soup.find_all('link'):
        all_css.append(css_tag)
        css_tag.extract()
    return render_with_plugin('plugin.html', current_plugin={'name': name},
                              plugin_content=str(soup),
                              plugin_javascript=all_scripts,
                              plugin_css=all_css)


@ui.route('/settings')
def settings():
    report_handler.page_in('settings')
    return render_with_plugin('settings.html')


@ui.route('/group_list')
def group_list():
    groups = context.application.data_manager.data_groups
    current_group = context.application.data_manager.current_data_group
    current_group_name = None
    if current_group:
        current_group_name = os.path.basename(current_group.dir_path)
    return render_template(

        'dm_v1/data_group_list.html',
        groups=groups.keys(),
        current_group=current_group_name
    )


@ui.route('/group_btn/')
@ui.route('/group_btn/<string:name>')
def group_btn(name=None):
    current_group = context.application.data_manager.current_data_group
    current_group_name = None
    if current_group:
        current_group_name = current_group.name
    if not name or name == '':
        return render_template(
            'dm_v1/data_group_btn.html',
            group=None,
            current_group=current_group_name)
    if context.application.data_manager.data_groups.get(name):
        return render_template(
            'dm_v1/data_group_btn.html',
            group=name,
            current_group=current_group_name)
    return context.make_fail_response('Group not found')



import json
from queue import Queue
from concurrent.futures import ThreadPoolExecutor
import requests
from lyrebird.mock import context
import datetime
from ...version import VERSION
import subprocess
import sys
import getpass
import os
from .. import plugin_manager
from lyrebird import log


lyrebird_start_time = None
last_page = None
last_page_in_time = None


logger = log.get_logger()


class ReportHandler:
    """
    用于上报到ELK的处理器

    """

    def __init__(self):
        self.report_url = None
        self.report_headers = None
        self.event_executor = ThreadPoolExecutor(max_workers=5)

    def report_worker(self, data):
        # 如果配置文件无reporter的配置
        conf = context.application.conf
        if not conf.get('reporter.url') or not conf.get('reporter.headers'):
            return
        
        self.report_url = conf['reporter.url']
        self.report_headers = conf['reporter.headers']
        
        if isinstance(data, dict):
            data.update(base_data)
        
        try:
            requests.request("POST", self.report_url, json=data, headers=self.report_headers)
        except Exception as e:
            logger.error(f'Send report failed. {e}')

    def send_report(self, data):
        self.event_executor.submit(self.report_worker, data)


report_handler = ReportHandler()


def report(data):
    report_handler.send_report(data)


def _env_info():
    user_name = subprocess.run('git config user.name', shell=True, stdout=subprocess.PIPE).stdout.decode().strip()
    user_email = subprocess.run('git config user.email', shell=True, stdout=subprocess.PIPE).stdout.decode().strip()
    return {
        'username': getpass.getuser(),
        'platform': sys.platform,
        'version': sys.version,
        'argv': sys.argv,
        'git.user_name': user_name, 
        'git.user_email': user_email,
        'pid': os.getpid()
        }

def _make_base_data():
        # web plugin 信息
    web_plugins = {}
    for name in plugin_manager.web_plugins:
        plugin = plugin_manager.web_plugins[name]
        web_plugins[name] = {
            'class': plugin.__class__.__name__,
            'module': plugin.__module__
        }
    # data plugin 信息
    data_plugins = {}
    for name in plugin_manager.data_handler_plugins:
        plugin = plugin_manager.data_handler_plugins[name]
        data_plugins[name] = {
            'class': plugin.__class__.__name__,
            'module': plugin.__module__
        }
    data = {
        'reporter':'lyrebird', 
        'version':VERSION, 
        'webPlugins': web_plugins,
        'dataPlugins': data_plugins, 
        'env': _env_info()
        }

    return data

    
base_data = _make_base_data()


def _make_data(action):
    data = {}
    data.update(base_data)
    data['action'] = action
    data['time'] = str(datetime.datetime.now())
    return data


def _page_out():
    global last_page
    global last_page_in_time

    if last_page and last_page_in_time:
        duration = datetime.datetime.now() - last_page_in_time
        page_out_data = _make_data('page.out')
        page_out_data['page'] = last_page
        page_out_data['duration'] = duration.total_seconds()
        report(page_out_data)


def page_in(name):
    _page_out()
    
    global last_page
    global last_page_in_time

    data = _make_data('page.in')
    data['page'] = name
    report(data)
    
    last_page = name
    last_page_in_time = datetime.datetime.now()


def start():
    global lyrebird_start_time
    lyrebird_start_time = datetime.datetime.now()
    data = _make_data('start')
    report(data)


def stop():
    _page_out()
    data = _make_data('stop')
    data['duration'] = (datetime.datetime.now() - lyrebird_start_time).total_seconds()
    report(data)


from .data import Flow
from pathlib import Path
import codecs
import json
from .exceptions import GroupNotFoundError



class Group:
    
    def __init__(self, path: Path):
        self.path = path
        self.name = self.path.name
        self.data_array = []
        self.request_map = {}
    
    def save(self):
        # create group dir
        if not self.path.exists():
            self.path.mkdir()
        # update map file

        # update data
        for data in self.data_array:
            data.save()
    
    def load(self):
        if not self.path.exists():
            raise GroupNotFoundError(f'Load group failed. Group {self.path} not found.')
        for sub_file in self.path.iterdir():
            if sub_file.is_file() and sub_file.name == 'conf.json':
                with codecs.open(str(sub_file) , 'r', 'utf-8') as f:
                    self.request_map = json.loads(f.read())
            elif sub_file.is_dir():
                data = Flow.from_path(sub_file)
                self.data_array.append(data)

    def add_data(self, data):
        pass

    def delete_data(self, data):
        pass

    def find_data(self, url, **kwargs):
        pass



from flask import Request, Response
from .. import context
from lyrebird import application
import uuid
import time


class HandlerContext:
    """
    请求处理器上下文变量
    用于保存一个请求处理过程中的request, response

    """

    def __init__(self):
        self.id = str(uuid.uuid4())
        self.request: Request = None
        self.response: Response = None
        self.client_req_time = None
        self.client_resp_time = None
        self.server_req_time = None
        self.server_resp_time = None
        self.flow = {'id': self.id}

    def update_client_req_time(self):
        self.client_req_time = time.time()
        self.request2dict()
        # 消息总线 客户端请求事件
        context.application.event_bus.publish('flow',
                                              dict(name='client.request',
                                              time=self.client_req_time,
                                              id=self.id,
                                              flow=self.flow
                                              ))

    def update_client_resp_time(self):
        self.client_resp_time = time.time()
        self.response2dict()
        # 消息总线 客户端响应事件
        context.application.event_bus.publish('flow',
                                              dict(name='client.response',
                                              time=self.client_resp_time,
                                              id=self.id,
                                              flow=self.flow))

    def update_server_req_time(self):
        self.server_req_time = time.time()
        self.request2dict()
        # 消息总线 客户端请求事件
        context.application.event_bus.publish('flow',
                                              dict(name='server.request',
                                              time=self.server_req_time,
                                              id=self.id,
                                              flow=self.flow))

    def update_server_resp_time(self):
        self.server_resp_time = time.time()
        self.response2dict()
        # 消息总线 客户端请求事件
        context.application.event_bus.publish('flow',
                                              dict(name='server.response',
                                              time=self.server_resp_time,
                                              id=self.id,
                                              flow=self.flow))

    def get_origin_url(self):
        proxy_headers = application.config['mock.proxy_headers']
        scheme = self.request.headers.get(proxy_headers['scheme'], default='http')
        host = self.request.headers.get(proxy_headers['host'])
        port = self.request.headers.get(proxy_headers['port'], default='80')
        url_index = self.request.url.find('/mock/')
        if scheme and host and port:
            scheme = scheme.strip()
            host = host.strip()
            port = port.strip()
            if port == '80':
                url = scheme+'://'+host
            else:
                url = scheme+'://'+host+':'+port
            if self.request.path and len(self.request.path[len('/mock'):]) > 0:
                url = url + self.request.path[len('/mock'):]
            query_string = self.request.query_string.decode()
            if query_string:
                url = url + '?' + query_string
            return url
        elif url_index > 0:
            return self.request.url[url_index + 6:]
        else:
            return None

    def request2dict(self):
        if self.request is None:
            return
        req = {}
        req['url'] = self.request.url
        req['method'] = self.request.method
        req['headers'] = [{'name':header[0], 'value':header[1]} for header in self.request.headers]
        self.flow['request'] = req

    def response2dict(self):
        if self.response is None:
            return
        resp = {}
        resp['status'] = self.response.status_code
        resp['headers'] = [{'name': header[0], 'value': header[1]} for header in self.response.headers]
        self.flow['response'] = resp


from collections import deque


_cache = None


def get_cache():
    global _cache
    if not _cache:
        # todo 此处应根据配置，决定生成ListCache还是RedisCache
        _cache = ListCache()
    return _cache


class ListCache:
    """
    双向序列
    默认最大值1000
    存储流经mock服务的数据
    
    """
    def __init__(self, maxlen=1000):
        self._cache = deque(maxlen=maxlen)

    def add(self, obj):
        self._cache.append(obj)

    def items(self):
        return list(self._cache)

    def clear(self):
        self._cache.clear()
    
    def delete(self, obj):
        self._cache.remove(obj)
    
    def delete_by_ids(self, *ids):
        del_items = []
        for item in list(self._cache):
            if item.id in ids:
                del_items.append(item)
        for item in del_items:
            self.delete(item)

class RedisCache:
    """
    如果部署在服务器上，并使用多进程，需要使用redis存储数据，实现多进程共享数据
    
    """
    pass


class FileCache:
    pass
